#!/bin/sh
###############################################################################
# This script runs start_encrypted_instance.sh in various combinations.       #
#                                                                             #
#     Example: ./test_encroot.sh trivial example.com                          #
#                                                                             #
# Have a look at the test_suites.sh file for all details on the test suites.  #
# EC2 Regions, build platforms, target systems, processor architecture, and   #
# options are looped over to confirm that everything is working as expected.  #
# Before running, you must configure your DNS with a separate subdomain for   #
# each tested region. The domain name should point to an Elastic IP address   #
# that is available within the region. Subdomains must named be as follows:   #
#                                                                             #
#     ap-northeast-1    Tokyo                  t1.example.com                 #
#     ap-southeast-1    Singapore              s1.example.com                 #
#     ap-southeast-2    Sydney                 s2.example.com                 #
#     eu-west-1         Ireland                i1.example.com                 #
#     sa-east-1         São Paulo              s3.example.com                 #
#     us-east-1         Northern Virginia      v1.example.com                 #
#     us-west-1         Northern California    c1.example.com                 #
#     us-west-2         Oregon                 o1.example.com                 #
#                                                                             #
# The boot.crt and boot.key files must be stored under ../SSL/ in a directory #
# named after the subdomain of each tested region (v1, t1, etc.) and any root #
# certificate should be saved in the ../SSL/root.crt file. The latter is only #
# needed if you made your own internal Certification Authority (See SSL.txt). #
#                                                                             #
###############################################################################
#                                                                             #
# Copyright (c) 2013 Henrik Gulbrandsen <henrik@gulbra.net>                   #
#                                                                             #
# This software is provided 'as-is', without any express or implied warranty. #
# In no event will the authors be held liable for any damages arising from    #
# the use of this software.                                                   #
#                                                                             #
# Permission is granted to anyone to use this software for any purpose,       #
# including commercial applications, and to alter it and redistribute it      #
# freely, subject to the following restrictions:                              #
#                                                                             #
# 1. The origin of this software must not be misrepresented; you must not     #
#    claim that you wrote the original software. If you use this software     #
#    in a product, an acknowledgment in the product documentation would be    #
#    appreciated but is not required.                                         #
#                                                                             #
# 2. Altered source versions must be plainly marked as such, and must not be  #
#    misrepresented as being the original software.                           #
#                                                                             #
# 3. This notice may not be removed or altered from any source distribution,  #
#    except that more "Copyright (c)" lines may be added to already existing  #
#    "Copyright (c)" lines if you have modified the software and wish to make #
#    your changes available under the same license as the original software.  #
#                                                                             #
###############################################################################

set -e

cd "$(dirname $0)"
. ./test_suites.sh

export AWSAPI_FAILURE_COMMAND="exit 1"
export EC2_API_VERSION="2013-02-01"
PATH="$(pwd):$PATH"

### Constants #################################################################

# Files copied to the build system
BUILD_FILES="activate.cgi awsapi boot.crt boot.key cryptsetup cryptsetup.sh
    hiding.gif index.html init.sh make_bozo_dir.sh make_encrypted_ubuntu.sh
    pre_init.sh start_encrypted_instance.sh uecimage.gpg"

# Output format and file
FORMAT="%-16s %-10s %-10s %-7s %-12s %s\n"
LOG_FILE="test_results.txt"

# Security group and SSH keys
GROUP="Encroot Test"
KEY_NAME="TEST_key1"
KEY_FILE="$HOME/.ssh/$KEY_NAME"

# Random passwords for the encryption test
PASSWORD1="aae856476490bf1466bcc8fecf38c1ac"
PASSWORD2="4951649e0063075ecbc75306a5f2783c"

# Let's hope this doesn't change...
UBUNTU_AWS_ACCOUNT="099720109477"

### Global Variables ##########################################################

base_domain=""
buildInstance_instanceId=""
buildInstance_ipAddress=""
current_platform=""
current_region=""
suite_name=""

### Parameters ################################################################

print_test_count() {
    local list count=1

    for list in REGIONS PLATFORMS TARGETS ARCHS OPTIONS; do
        eval set \$$list; count=$((count*$#));
    done

    echo $count;
}

if [ $# -lt 1 ] || [ $# -gt 2 ]; then
    echo "Usage: ${0##*/} [<suite>] <base-domain>"
    echo
    echo "Suites:"
    for suite in $SUITES; do
        DESCRIPTION="No description"
        SUITE_${suite}; count="$(print_test_count)"
        printf "    %-8s %-40s  %3d test" "$suite:" "$DESCRIPTION" $count
        if [ $count -eq 1 ]; then echo ""; else echo "s"; fi
    done
    echo
    exit 1
fi

### Validation ################################################################

if [ $# -eq 1 ]; then
    suite_name="default"
    base_domain="$1"
else
    suite_name="$1"
    base_domain="$2"
fi

if [ "${base_domain}" = "example.com" ] ||
   [ "${base_domain%.example.com}" != "${base_domain}" ]; then
    echo "No, dummy! You should use your own domain, not example.com!"
    exit 1
fi

if [ -z "${suite_name}" ]; then
    echo "Error: Blank suite"
    exit 1
fi

if [ -z "${base_domain}" ]; then
    echo "Error: Blank domain"
    exit 1
fi

if ! type "SUITE_${suite_name}" 2> /dev/null | grep -q "function"; then
    echo "Invalid suite: ${suite_name}"
    exit 1
fi

SUITE_${suite_name};

### Functions #################################################################

dots() {
    perl -e 'print $ARGV[0], "."x(40-length($ARGV[0])), "... "' "$*";
}

print_separator() {
    printf "%s" "----------------------------------------"
    echo "---------------------------------------"
}

log() {
    printf "$@" >> "$LOG_FILE";
}

need() {
    if [ ! -e "$1" ]; then
        echo "Missing file: $1 - $2"
        exit 1
    fi
}

print_domain_for_region() {
    local region=$1
    local domain

    case $region in
        ap-northeast-1) domain="t1.$base_domain" ;; # Tokyo
        ap-southeast-1) domain="s1.$base_domain" ;; # Singapore
        ap-southeast-2) domain="s2.$base_domain" ;; # Sydney
        eu-west-1)      domain="i1.$base_domain" ;; # Ireland
        sa-east-1)      domain="s3.$base_domain" ;; # São Paulo
        us-east-1)      domain="v1.$base_domain" ;; # Northern Virginia
        us-west-1)      domain="c1.$base_domain" ;; # Northern California
        us-west-2)      domain="o1.$base_domain" ;; # Oregon
    esac

    echo "$domain"
}

print_instance_name() {
    local option="$1" target="$2" arch="$3"
    local name

    case $option in
      --big-boot) name="TEST_BigBoot";;
      *) name="TEST_Normal";;
    esac

    name="$name$(echo $target | perl -pe 's/(.*)/\u$1/')"

    case $arch in
      i386) name="${name}32";;
      x86_64) name="${name}64";;
    esac

    echo "$name"
}

confirm_ssh() {
    local address="$1"
    local known attempt

    # Make sure that the address won't make SSH nervous
    known="$HOME/.ssh/known_hosts"
    grep -v "^$address " "${known}" 2> /dev/null > "${known}.new"
    mv "${known}.new" "${known}"

    # Wait for the SSH daemon to start
    for attempt in 0 1 2 3 4 5 6 7 8 9; do sleep 10
        ssh -i "$KEY_FILE" -o StrictHostKeyChecking=no "ubuntu@$address" true \
            2>/dev/null && break || true # Ignore errors; break on success
        if [ $attempt = 9 ]; then
            return 1
        fi
    done

    return 0
}

activate_server() {
    local domain="$1"
    local curlOptions

    curlOptions="--data key=$PASSWORD1"
    if [ -e "../SSL/root.crt" ]; then
        curlOptions="$curlOptions --cacert ../SSL/root.crt";
    fi

    curlOptions="$curlOptions --silent --show-error"
    curlOptions="$curlOptions --output /dev/null"
    curlOptions="$curlOptions --connect-timeout 5"
    curlOptions="$curlOptions --retry 12"

    curl $curlOptions "https://$domain/cgi-bin/activate.cgi" || return 1

    return 0;
}

terminate_domain() {
    local domain="$1"
    local ipAddress instance instanceList

    # Find the IP address of the domain
    ipAddress="$(dig +short "$domain")"
    if [ -z "$ipAddress" ]; then
        echo "No IP address for $domain"
        exit 1
    fi

    # Find test instances with the given IP address
    $(awsapi ec2.DescribeInstances instance+reservationSet.n.instancesSet.1.{ \
        instanceId, keyName eq "$KEY_NAME", \
        group:groupSet.1.groupName eq "$GROUP", \
        state:instanceState.name ne terminated, \
        ipAddress eq "$ipAddress" \
    })

    # We only expect one instance to match, but...
    for instance in $instanceList; do

        # Confirm the address, just in case our awsapi filter is buggy
        if [ "$(instance.ipAddress)" != "$ipAddress" ]; then
            echo "Bad IP address: \"$(instance.ipAddress)\" != \"$ipAddress\""
            exit 1
        fi

        # Confirm the key, in case the bug is somewhere else
        if [ "$(instance.keyName)" != "$KEY_NAME" ]; then
            echo "Wrong key: $(instance.keyName)";
            exit 1
        fi

        # Check the group too, just to be sure
        if [ "$(instance.group)" != "$GROUP" ]; then
            echo "Wrong group: $(instance.group)";
            exit 1
        fi

        # If everything looks OK, terminate the instance
        $(awsapi ec2.TerminateInstances \
            InstanceId.1="$(instance.instanceId)")

    done
}

prepare_region() {
    local region="$1"
    local query

    # Set the endpoint for the new region
    export EC2_ENDPOINT="https://ec2.$region.amazonaws.com/"

    # Create and import a separate key pair if needed
    query="awsapi ec2.DescribeKeyPairs KeyName.1"
    if $query="$KEY_NAME" 2>&1 | grep -q InvalidKeyPair.NotFound; then
        dots "Importing key pair"
        if [ ! -e "$KEY_FILE" ]; then
            mkdir -p "$(dirname "$KEY_FILE")"
            ssh-keygen -q -t rsa -b 2048 -N "" -C "$KEY_NAME" -f "$KEY_FILE"
        fi
        $(awsapi ec2.ImportKeyPair KeyName="$KEY_NAME" PublicKeyMaterial="$(
            cat "$KEY_FILE".pub | openssl base64 | tr -d '\n')")
        echo "done"
    fi

    # Confirm that we have the private key
    if [ ! -e "$KEY_FILE" ]; then
        echo "Missing key: $KEY_FILE"
        exit 1
    fi

    # Is the security group missing?
    query="awsapi ec2.DescribeSecurityGroups GroupName.1"
    if $query="$GROUP" 2>&1 | grep -q InvalidGroup.NotFound; then
        dots "Creating security group..."

        # Then create a new security group...
        $(awsapi ec2.CreateSecurityGroup GroupName="$GROUP" \
            GroupDescription="Used by ${0##*/}")

        # ...and add SSH and HTTPS ingress rules
        $(awsapi ec2.AuthorizeSecurityGroupIngress \
            GroupName="$GROUP" \
            IpPermissions.1.{ \
                IpProtocol=tcp, \
                FromPort=22, ToPort=22, \
                IpRanges.1.CidrIp="0.0.0.0/0" \
            } \
            IpPermissions.2.{ \
                IpProtocol=tcp, \
                FromPort=443, ToPort=443, \
                IpRanges.1.CidrIp="0.0.0.0/0" \
            } \
        )

        echo "done"
    fi
}

prepare_build_system() {
    local region="$1" platform="$2"

    # Kill the old instance
    if [ -n "$buildInstance_instanceId" ]; then
        $(awsapi ec2.TerminateInstances \
            InstanceId.1="$buildInstance_instanceId")
    fi

    # Prepare the region environment
    if [ "$region" != "$current_region" ]; then
        prepare_region "$region"
        current_region="$region"
    fi

    # Create a new instance
    prepare_build_instance "$region" "$platform" "x86_64"
}

prepare_build_instance() {
    local region="$1" platform="$2" arch="$3"
    local object objectList
    local imageId bits domain

    # Find a suitable AMI for this region, platform, and architecture
    dots "Selecting image"
    imageId=$($(awsapi --table ec2.DescribeImages Owner.1=$UBUNTU_AWS_ACCOUNT \
        Filter.1.Name=architecture Filter.1.Value.1=$arch \
        Filter.2.Name=root-device-type Filter.2.Value.1=ebs \
        object+imagesSet.n.{name,imageId}) | \
        grep ubuntu/images/ebs/ubuntu-$platform | \
        sort | tail -1 | awk '{print $2}')
        echo "$imageId"

    # Complain if it doesn't look like an image ID
    if [ "${imageId}" = "${imageId#ami-}" ]; then
        echo "Weird AMI: \"$imageId\""
        exit 1
    fi

    # Start the instance and install necessary files
    start_build_instance "$imageId" "$platform" "$arch"
    print_separator; install_files "$region"
}

start_build_instance() {
    local imageId="$1" platform="$2" arch="$3"
    local instanceId bits state ipAddress

    # Start the new build instance
    dots "Starting instance"
    $(awsapi ec2.RunInstances ImageId="$imageId" MinCount=1 MaxCount=1 \
        KeyName="$KEY_NAME" SecurityGroup.1="$GROUP" InstanceType=t1.micro \
        instancesSet.1.instanceId)
        sleep 5; echo "$instanceId"

    # Give it a name like "TEST_Precise64_Build"
    dots "Naming instance"
    if [ "$arch" = i386 ]; then bits=32; else bits=64; fi
    name="$name$(echo $platform | perl -pe 's/(.*)/\u$1/')"
    $(awsapi ec2.CreateTags ResourceId.1="$instanceId" \
        Tag.1.{ Key=Name, Value="TEST_${name}${bits}_Build" })
        echo "TEST_${name}${bits}_Build"

    # Wait for the new instance to start
    dots "Waiting for instance"
    $(awsapi ec2.DescribeInstances \
        Filter.1.{ Name="instance-id", Value.1="$instanceId" } \
        reservationSet.1.instancesSet.1.{ \
            state:instanceState.name := -/pending/running, \
            ipAddress \
        }); echo "done"

    # Wait for the SSH daemon to start
    dots "Waiting for SSH daemon"
    if ! confirm_ssh $ipAddress; then
        echo "Could not reach $ipAddress"
        exit 1
    fi

    # Wait and repeat, since SSH host keys may change
    sleep 10; if ! confirm_ssh $ipAddress; then
        echo "Lost connection to $ipAddress"
        exit 1
    fi

    # Save the necessary details
    buildInstance_instanceId="$instanceId"
    buildInstance_ipAddress="$ipAddress"
    echo "done";
}

install_files() {
    local region="$1"
    local ipAddress="$buildInstance_ipAddress"
    local domain sslDir ttyState

    # Select certificates for the domain
    domain=$(print_domain_for_region "$region")
    sslDir="../SSL/${domain%.$base_domain}"
    cp "$sslDir/boot.crt" "$sslDir/boot.key" .

    # Make the next step look nicer
    ttyState=$(stty -g)
    stty cols 80

    # Copy all necessary files
    ssh -i "$KEY_FILE" "ubuntu@$ipAddress" mkdir encroot
    scp -i "$KEY_FILE" $BUILD_FILES "ubuntu@$ipAddress:encroot"
    scp -i "$KEY_FILE" "$HOME"/.awsapirc "ubuntu@$ipAddress:/home/ubuntu"
    stty "$ttyState"
    print_separator;

    # Install cryptsetup
    ssh -ti "$KEY_FILE" "ubuntu@$ipAddress" sudo apt-get -y install cryptsetup
    print_separator;

    # These should not be packaged...
    rm -f boot.crt boot.key
}

test_combination() {
    local region="$1" platform="$2" target="$3" arch="$4" option="${5#none}"
    local domain name args

    print_separator;

    # Prepare everything for the main encroot script
    if [ "${region}_${platform}" != "$current_platform" ]; then
        prepare_build_system "$region" "$platform";
        current_platform="${region}_${platform}"
    fi

    # Prepare script arguments
    domain=$(print_domain_for_region "$region")
    name=$(print_instance_name "$option" "$target" "$arch")
    option="${option:+ $option}"

    # Adjust target for 32/64-bit
    case $arch in
      i386) target="${target}/i386";;
      x86_64) target="${target}/amd64";;
    esac

    # Run the encroot scripts
    args="--name='$name'$option --system='$target' --group='$GROUP' '$domain'"
    printf "$PASSWORD1\n$PASSWORD1\n$PASSWORD2\n$PASSWORD2\n\n" | \
      ssh -tt -i "$KEY_FILE" "ubuntu@$buildInstance_ipAddress" \
        "encroot/start_encrypted_instance.sh $args" \
            || return 1

    # Activate the new server
    dots "Activating server"
    if ! activate_server "$domain"; then
        echo "failed"; terminate_domain "$domain"; return 1;
    else
        echo "done"
    fi

    # Verify that it comes up
    dots "Confirming SSH"
    if ! confirm_ssh "$domain"; then
        echo "failed"; terminate_domain "$domain"; return 1;
    else
        echo "done"
    fi

    # Terminate the server
    terminate_domain "$domain"
    return 0;
}

### Sanity Check ##############################################################

for region in $REGIONS; do

    # Verify the DNS configuration
    domain=$(print_domain_for_region "$region")
    if [ -z $(dig +short "$domain") ]; then
        echo "Missing domain for $region: $domain"
        exit 1
    fi

    # Confirm that we have necessary files
    dir="../SSL/${domain%.$base_domain}"
    need "$dir/boot.crt" "SSL certificate for $domain"
    need "$dir/boot.key" "private SSL key for $domain"

done

### Run the tests #############################################################

combinations=""

# Make a list of all combinations
for region in $REGIONS; do
  for platform in $PLATFORMS; do
    for target in $TARGETS; do
      for arch in $ARCHS; do
        for option in $OPTIONS; do
          combinations="$combinations $region:$platform:$target:$arch:$option"
        done
      done
    done
  done
done

# Print the result header
:> $LOG_FILE # ...starting with an empty file
log "$FORMAT" "Region" "Platform" "Target" "Arch" "Option" "Result"
log "%s\n" "$(print_separator)"

# Test all combinations
for combination in $combinations; do
    combination=$(echo $combination | sed 's/:/ /g')
    if test_combination $combination; then result="OK"; else result="Bad"; fi
    log "$FORMAT" $combination $result
done

# Terminate the remaining build instance
if [ -n "$buildInstance_instanceId" ]; then
    $(awsapi ec2.TerminateInstances InstanceId.1="$buildInstance_instanceId")
fi

###############################################################################
